#!/usr/bin/env python
# flake8: noqa: E402

"""
A simple load tester for IRRd.
Sends random !g queries.
"""
import time

import argparse
import random
import socket


def main(host, port, count):
    queries = [b'!!\n']
    for i in range(count):
        asn = random.randrange(1, 50000)
        query = f'!gAS{asn}\n'.encode('ascii')
        queries.append(query)
    queries.append(b'!q\n')

    s = socket.socket()
    s.settimeout(3000)
    s.connect((host, port))

    queries = b"""!!
!aAS-10021ARCSTAR
!aAS-10099NTT
!aAS-10204
!aAS-11114
!aAS-131248-CUST
!aAS-132132-TRANSIT
!aAS-132876
!aAS-133045-TRANSIT
!aAS-134842
!aAS-135377
!aAS-16524
!aAS-17534
!aAS-17666
!aAS-17958
!aAS-18149ARCSTAR
!aAS-18683-MWFN-TRANSIT
!aAS-19016
!aAS-19303
!aAS-1AND1
!aAS-20264
!aAS-20394-CUSTOMERS
!aAS-21VIANET
!aAS-23655-ALL
!aAS-24413
!aAS-2500-2914
!aAS-25766-DOWNSTREAM
!aAS-26167
!aAS-29802
!aAS-30517-ASSOCIATED
!aAS-30640
!aAS-31515ARCSTAR
!aAS-33606
!aAS-35916-TRANSIT
!aAS-38566
!aAS-395412-TRANSIT
!aAS-395662
!aAS-3TCOM
!aAS-3W-INFRA
!aAS-40676
!aAS-45903
!aAS-46043
!aAS-4628
!aAS-46475
!aAS-4648-Customers
!aAS-46652
!aAS-46844
!aAS-4693ARCSTAR
!aAS-4761
!aAS-48196
!aAS-4D-DC
!aAS-50010
!aAS-5009
!aAS-52551
!aAS-55532ARCSTAR
!aAS-55805
!aAS-55967
!aAS-5713
!aAS-57344
!aAS-58463
!aAS-58682
!aAS-59318-AS-KH
!aAS-60917
!aAS-61832-clientes
!aAS-62833-ny4
!aAS-63916
!aAS-7552
!aAS-7562
!aAS-9268ARCSTAR
!aAS-9381
!aAS-9416-TRANSIT
!aAS-9908VERIO
!aAS-9929ARCSTAR
!aAS-A2B
!aAS-AA
!aAS-AAMRA-ATL-BD
!aAS-ACCELERATED
!aAS-ACCESSFORALL
!aAS-ACCRETIVE
!aAS-ACESSO10
!aAS-ACORUS
!aAS-ADAM-EU
!aAS-ADAMOEU
!aAS-ADAMS
!aAS-ADHOST-PEERING
!aAS-ADNC-SAN
!aAS-ADNET
!aAS-ADNET-DC
!aAS-ADNET-DC6
!aAS-ADNET6
!aAS-AEG
!aAS-AFILIAS-FULLMONTY
!aAS-AGARTO
!aAS-AGARTO6
!aAS-AGILIT
!aAS-AGODA
!aAS-AI
!aAS-AIE
!aAS-AIRBANDUK
!aAS-AIRBANDUK-IPV6
!aAS-AIRWIRE
!aAS-AKAMAI
!aAS-ALCOM
!aAS-ALHAMBRA
!aAS-ALLIED_TELECOM
!aAS-ALTECOM
!aAS-AMATIS
!aAS-AMAZON
!aAS-AMC
!aAS-AMCL-BB
!aAS-AMERICANET-BR
!aAS-AMERIDOTCA
!aAS-AMERIDOTCA6
!aAS-AMPERSAND
!aAS-ANEXIA
!aAS-ANIN
!aAS-ANJOS
!aAS-ANTEL-ALL
!aAS-ANX
!aAS-ANYCAST
!aAS-AOFEI-Intl
!aAS-AOI
!aAS-AORTA
!aAS-AORTA6
!aAS-APPLE
!aAS-APPNOR
!aAS-ARENAONE
!aAS-ARI
!aAS-ARIN-NTT-SILICON-VALLEY
!aAS-ARIN-SERVICE-EAST
!aAS-ARMSTRONGPXY
!aAS-ARPNETWORKS
!aAS-AS29550
!aAS-AS29550-V6
!aAS-ASAHI-NET
!aAS-ASK4
!aAS-ASTRALUS
!aAS-ATOM86
!aAS-AUTOMATTIC
!aAS-AVIRADE
!aAS-AXNE
!aAS-AYERA
!aAS-AZERSATMEMBERS
!aAS-BANDWIDTH
!aAS-BBC
!aAS-BBCONNECT
!aAS-BBOI
!aAS-BDRV
!aAS-BEANFIELD
!aAS-BELBONEMEMBERS
!aAS-BELCLOUD
!aAS-BELCLOUD-V6
!aAS-BELNET
!aAS-BENDTEL
!aAS-BERTRAMWIRELESS
!aAS-BESTPATH
!aAS-BGCOM
!aAS-BGL
!aAS-BGPNET
!aAS-BIDSTRADING
!aAS-BIGPIPE
!aAS-BIH
!aAS-BINCNET
!aAS-BIT
!aAS-BIT6
!aAS-BKNIX
!aAS-BKVG
!aAS-BLACKNIGHT
!aAS-BLIZZARD
!aAS-BLRS
!aAS-BOGONS
!aAS-BORWOOD
!aAS-BOUYGTEL-ISP
!aAS-BRIDGEFIBRE
!aAS-BROADASPECT
!aAS-BSKYB-BROADBAND
!aAS-BT-GLOBAL
!aAS-BTC-UP
!aAS-BTC-UP-V6
!aAS-BTELOUT
!aAS-BTELOUT-V6
!aAS-BTS
!aAS-BTS-GLOBAL
!aAS-BULSATCOM
!aAS-BURSTFIRE-NTT
!aAS-BWFN
!aAS-BXI
!aAS-BYTEMARK
!aAS-BatBlue
!aAS-C4L
!aAS-CACHENETWORKS
!aAS-CADENCE
!aAS-CAIS
!aAS-CALL27
!aAS-CAMBRIUM
!aAS-CAMPAIGNMONITOR
!aAS-CANAR-TRANSIT
!aAS-CAPENET
!aAS-CAPITALONLINEDATAgit 
!aAS-CARPATHIA
!aAS-CATALYST2
!aAS-CATALYST2-V6
!aAS-CAVEO
!aAS-CBCCOM
!aAS-CBJAYABBAND
!aAS-CCCN
!aAS-CCI
!aAS-CCI-FL-CUST
!aAS-CCI-KS-CUST
!aAS-CCSLEEDS
!aAS-CDNETWORKS
!aAS-CELESTE
!aAS-CENTERPRESTADORA
!aAS-CENTRILOGIC
!aAS-CESNET
!aAS-CESNET6
!aAS-CETIN
!aAS-CHAOS
!aAS-CHIEF-TRANSIT
!aAS-CHOOPA
!aAS-CIFNET
!aAS-CIPHERKEY-TRANSIT
!aAS-CITIZENSCOOP
!aAS-CJH
!aAS-CLEARIP
!aAS-CLOUDFLARE
!aAS-CLOUDHELIX
!aAS-CLOUDMERGE
!aAS-CLOUVIDER
!aAS-CMI-ALL
!aAS-CN2
!aAS-CNA
!aAS-CNHAT
!aAS-CNHAT-V6
!aAS-CODETEL
!aAS-COLO-AU
!aAS-COLOAT
!aAS-COLOCLUE
!aAS-COLT
!aAS-COLT6
!aAS-COMSAVE
!aAS-COMVIVE
!aAS-CONNETU
!aAS-CONTINENT8-ASIA
!aAS-CORALNET
!aAS-COREIX
!aAS-CORESPACE-DAL
!aAS-COREXCHANGE-US
!aAS-CORPORATIVATELCOM
!aAS-COX-TRANSIT
!aAS-CPNET
!aAS-CPRM
!aAS-CRITEO-AP
!aAS-CSF
!aAS-CSLOXINFO-TRANSIT
!aAS-CTCX
!aAS-CTI-Telecom
!aAS-CTMTELEONE
!aAS-CUSTDC
!q""".splitlines()

    queries_str = b'\n'.join(queries) + b'\n'
    s.sendall(queries_str)

    start_time = time.perf_counter()
    while 1:
        data = s.recv(1024*1024)
        if not data:
            break

    count = len(queries) - 1
    elapsed = time.perf_counter() - start_time
    time_per_query = elapsed / count * 1000
    qps = int(count / elapsed)
    print(f'Ran {count} queries in {elapsed}s, time per query {time_per_query} ms, {qps} qps')


if __name__ == '__main__':  # pragma: no cover
    description = """A simple load tester for IRRd. Sends random !g queries."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--count', dest='count', type=int, default=5000,
                        help=f'number of queries to run (default: 5000)')
    parser.add_argument('host', type=str,
                        help='hostname of instance')
    parser.add_argument('port', type=int,
                        help='port of instance')
    args = parser.parse_args()

    main(args.host, args.port, args.count)



from typing import Tuple


def parse_as_number(value: str) -> Tuple[str, int]:
    """Validate and clean an AS number. Returns it in ASxxxx and numeric format."""
    value = value.upper()
    if not value.startswith('AS'):
        raise ValidationError(f'Invalid AS number {value}: must start with "AS"')

    if not value[2:].isnumeric():
        raise ValidationError(f'Invalid AS number {value}: number part is not numeric')

    value_int = int(value[2:])
    if value_int > 4294967295:
        raise ValidationError(f'Invalid AS number {value}: maximum value is 4294967295')

    return 'AS' + str(value_int), value_int


class ValidationError(ValueError):
    pass



import logging
from collections import defaultdict
from datetime import datetime, timezone
from typing import List, Set, Dict, Any, Tuple, Optional, Iterator, Union

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg

from irrd.conf import get_setting
from irrd.rpsl.parser import RPSLObject
from irrd.rpsl.rpsl_objects import OBJECT_CLASS_MAPPING
from .preload import get_preloader
from .queries import (BaseRPSLObjectDatabaseQuery, DatabaseStatusQuery,
                      RPSLDatabaseObjectStatisticsQuery)
from . import get_engine
from .models import RPSLDatabaseObject, RPSLDatabaseJournal, DatabaseOperation, RPSLDatabaseStatus

logger = logging.getLogger(__name__)
MAX_RECORDS_CACHE_BEFORE_INSERT = 5000


class DatabaseHandler:
    """
    Interface for other parts of IRRD to talk to the database.

    Note that no writes to the database are final until commit()
    has been called - and rollback() can be called at any time
    to all submitted changes.
    """
    journaling_enabled: bool
    _transaction: sa.engine.base.Transaction
    _rpsl_pk_source_seen: Set[str]
    # The RPSL upsert cache is a list of tuples. Each tuple first has a dict
    # with all database column names and their values, and then an optional int
    # with a forced serial. The forced serial is used for mirrored sources,
    # where this basically means: this call was triggered by an NRTM operation
    # with this serial.
    _rpsl_upsert_cache: List[Tuple[dict, Optional[int]]]

    def __init__(self, enable_preload_update=True):
        """
        Create a new database handler.

        If enable_preload is True (default), the Preloader will be notified
        of changed objects after every committed transaction. This should
        be disabled in tasks that do not run in the IRRd server process,
        such as the submit_email script.
        """
        self.journaling_enabled = True
        self._connection = get_engine().connect()
        self._start_transaction()
        self.preloader = None
        if enable_preload_update:
            self.preloader = get_preloader()

    def _start_transaction(self) -> None:
        """Start a fresh transaction."""
        self._transaction = self._connection.begin()
        self._rpsl_upsert_cache = []
        self._rpsl_pk_source_seen: Set[str] = set()
        self._object_classes_modified: Set[str] = set()
        self.status_tracker = DatabaseStatusTracker(self, journaling_enabled=self.journaling_enabled)

    def disable_journaling(self):
        self.journaling_enabled = False
        self.status_tracker.journaling_enabled = False

    def commit(self) -> None:
        """
        Commit any pending changes to the database and start a fresh transaction.
        """
        self._flush_rpsl_object_upsert_cache()
        self.status_tracker.finalise_transaction()
        try:
            self._transaction.commit()
            if self.preloader:
                self.preloader.reload(self._object_classes_modified)
            self._start_transaction()
        except Exception as exc:  # pragma: no cover
            self._transaction.rollback()
            logger.error('Exception occurred while committing changes, rolling back', exc_info=exc)
            raise

    def rollback(self) -> None:
        """Roll back the current transaction, discarding all submitted changes."""
        self._rpsl_upsert_cache = []
        self._rpsl_pk_source_seen = set()
        self._transaction.rollback()
        self._start_transaction()

    def execute_query(self, query: Union[BaseRPSLObjectDatabaseQuery, DatabaseStatusQuery, RPSLDatabaseObjectStatisticsQuery]) -> Iterator[Dict[str, Any]]:
        """Execute an RPSLDatabaseQuery within the current transaction."""
        # To be able to query objects that were just created, flush the cache.
        self._flush_rpsl_object_upsert_cache()
        statement = query.finalise_statement()
        result = self._connection.execute(statement)
        for row in result.fetchall():
            yield dict(row)
        result.close()

    def execute_statement(self, statement):
        """Execute a raw SQLAlchemy statement, without flushing the upsert cache."""
        return self._connection.execute(statement)

    def upsert_rpsl_object(self, rpsl_object: RPSLObject, forced_serial: Optional[int]=None) -> None:
        """
        Schedule an RPSLObject for insertion/updating.

        This method will insert the object, or overwrite an existing object
        if it has the same RPSL primary key and source. No other checks are
        applied before overwriting.

        Writes may not be issued to the database immediately for performance
        reasons, but commit() will ensure all writes are flushed to the DB first.

        The forced serial is needed for mirrored sources, where this basically means:
        this call was triggered by an NRTM operation with this serial.
        """
        ip_first = str(rpsl_object.ip_first) if rpsl_object.ip_first else None
        ip_last = str(rpsl_object.ip_last) if rpsl_object.ip_last else None

        ip_size = None
        if rpsl_object.ip_first and rpsl_object.ip_last:
            ip_size = rpsl_object.ip_last.int() - rpsl_object.ip_first.int() + 1

        # In some cases, multiple updates may be submitted for the same object.
        # PostgreSQL will not allow rows proposed for insertion to have duplicate
        # constrained values - so if a second object appears with a pk/source
        # seen before, the cache must be flushed right away, or the two updates
        # will conflict.
        source = rpsl_object.parsed_data['source']
        rpsl_pk_source = rpsl_object.pk() + '-' + source
        if rpsl_pk_source in self._rpsl_pk_source_seen:
            self._flush_rpsl_object_upsert_cache()

        self._rpsl_upsert_cache.append((
            {
                'rpsl_pk': rpsl_object.pk(),
                'source': source,
                'object_class': rpsl_object.rpsl_object_class,
                'parsed_data': rpsl_object.parsed_data,
                'object_text': rpsl_object.render_rpsl_text(),
                'ip_version': rpsl_object.ip_version(),
                'ip_first': ip_first,
                'ip_last': ip_last,
                'ip_size': ip_size,
                'asn_first': rpsl_object.asn_first,
                'asn_last': rpsl_object.asn_last,
                'updated': datetime.now(timezone.utc),
            },
            forced_serial,
        ))
        self._rpsl_pk_source_seen.add(rpsl_pk_source)
        self._object_classes_modified.add(rpsl_object.rpsl_object_class)

        if len(self._rpsl_upsert_cache) > MAX_RECORDS_CACHE_BEFORE_INSERT:
            self._flush_rpsl_object_upsert_cache()

    def delete_rpsl_object(self, rpsl_object: RPSLObject, forced_serial: Optional[int]=None) -> None:
        """
        Delete an RPSL object from the database.

        The forced serial is needed for mirrored sources, where this basically means:
        this call was triggered by an NRTM operation with this serial.
        """
        self._flush_rpsl_object_upsert_cache()
        table = RPSLDatabaseObject.__table__
        source = rpsl_object.parsed_data['source']
        stmt = table.delete(
            sa.and_(table.c.rpsl_pk == rpsl_object.pk(), table.c.source == source),
        ).returning(table.c.pk, table.c.rpsl_pk, table.c.source, table.c.object_class, table.c.object_text)
        results = self._connection.execute(stmt)

        if results.rowcount == 0:
            logger.error(f'Attempted to remove object {rpsl_object.pk()}/{source}, but no database row matched')
            return None
        if results.rowcount > 1:  # pragma: no cover
            # This should not be possible, as rpsl_pk/source are a composite unique value in the database scheme.
            # Therefore, a query should not be able to affect more than one row - and we also can not test this
            # scenario. Due to the possible harm of a bug in this area, we still check for it anyways.
            affected_pks = ','.join([r[0] for r in results.fetchall()])
            msg = f'Attempted to remove object {rpsl_object.pk()}/{source}, but multiple objects were affected, '
            msg += f'internal pks affected: {affected_pks}'
            logger.critical(msg)
            raise ValueError(msg)

        result = results.fetchone()
        self.status_tracker.record_operation(
            operation=DatabaseOperation.delete,
            rpsl_pk=result['rpsl_pk'],
            source=result['source'],
            object_class=result['object_class'],
            object_text=result['object_text'],
            forced_serial=forced_serial,
        )
        self._object_classes_modified.add(result['object_class'])

    def _flush_rpsl_object_upsert_cache(self) -> None:
        """
        Flush the current upsert cache to the database.

        This happens in one large INSERT .. ON CONFLICT DO UPDATE ..
        statement, which is more performant than individual queries
        in case of large datasets.
        """
        if not self._rpsl_upsert_cache:
            return

        rpsl_composite_key = ['rpsl_pk', 'source']
        stmt = pg.insert(RPSLDatabaseObject).values([x[0] for x in self._rpsl_upsert_cache])
        columns_to_update = {
            c.name: c
            for c in stmt.excluded
            if c.name not in rpsl_composite_key and c.name != 'pk'
        }

        update_stmt = stmt.on_conflict_do_update(
            index_elements=rpsl_composite_key,
            set_=columns_to_update,
        )

        try:
            self._connection.execute(update_stmt)
        except Exception as exc:  # pragma: no cover
            self._transaction.rollback()
            logger.error(f'Exception occurred while executing statement: {update_stmt}, rolling back', exc_info=exc)
            raise

        for obj, forced_serial in self._rpsl_upsert_cache:
            self.status_tracker.record_operation(
                operation=DatabaseOperation.add_or_update,
                rpsl_pk=obj['rpsl_pk'],
                source=obj['source'],
                object_class=obj['object_class'],
                object_text=obj['object_text'],
                forced_serial=forced_serial,
            )

        self._rpsl_upsert_cache = []
        self._rpsl_pk_source_seen = set()

    def delete_all_rpsl_objects_with_journal(self, source):
        """
        Delete all RPSL objects for a source from the database,
        all journal entries and the database status.
        This is intended for cases where a full re-import is done.
        Note that no journal records are kept of this change itself.
        """
        self._flush_rpsl_object_upsert_cache()
        table = RPSLDatabaseObject.__table__
        stmt = table.delete(table.c.source == source)
        self._connection.execute(stmt)
        table = RPSLDatabaseJournal.__table__
        stmt = table.delete(table.c.source == source)
        self._connection.execute(stmt)
        table = RPSLDatabaseStatus.__table__
        stmt = table.delete(table.c.source == source)
        self._connection.execute(stmt)
        # All objects are presumed to have been changed.
        self._object_classes_modified.update(OBJECT_CLASS_MAPPING.keys())

    def force_record_serial_seen(self, source: str, serial: int) -> None:
        """
        Forcibly record that a serial was seen for a source.

        This is used when receiving NRTM streams in which some objects are
        ignored or have errors. Because their operations are never saved, the
        serial needs to be manually advanced, to prevent re-querying the
        same objects all the time.
        """
        self.status_tracker.force_record_serial_seen(source, serial)

    def record_mirror_error(self, source: str, error: str) -> None:
        """
        Record an error seen in a mirrored database.
        Only the most recent error is stored in the DB status.
        """
        self.status_tracker.record_mirror_error(source, error)

    def record_serial_exported(self, source: str, serial: int) -> None:
        """
        Record an export of a source at a particular serial.
        """
        self.status_tracker.record_serial_exported(source, serial)

    def close(self) -> None:
        self._connection.close()


class DatabaseStatusTracker:
    """
    Keep track of the status of sources, and their journal, if enabled.
    Changes to the database should always call record_operation() on this object,
    and finalise_transaction() before committing.

    If journaling is enabled, a new entry in the journal will be made.
    If a journal entry was made, or forced_serial was set, a record is kept in
    memory with all serials encountered/created for this source.

    When finalising, the RPSLDatabaseStatus table is updated to correctly
    reflect the range of serials known for a particular source.
    """
    journaling_enabled: bool
    _new_serials_per_source: Dict[str, Set[int]]
    _sources_seen: Set[str]
    _mirroring_error: Dict[str, str]
    _exported_serials: Dict[str, int]

    c_journal = RPSLDatabaseJournal.__table__.c
    c_status = RPSLDatabaseStatus.__table__.c

    def __init__(self, database_handler: DatabaseHandler, journaling_enabled=True) -> None:
        self.database_handler = database_handler
        self.journaling_enabled = journaling_enabled
        self._reset()

    def force_record_serial_seen(self, source: str, serial: int):
        """
        Forcibly record that a serial was seen for a source.
        See DatabaseHandler.force_record_serial_seen for more info.
        """
        self._sources_seen.add(source)
        self._new_serials_per_source[source].add(serial)

    def record_mirror_error(self, source: str, error: str) -> None:
        """
        Record an error seen in a mirrored database.
        Only the most recent error is stored in the DB status.
        """
        self._sources_seen.add(source)
        self._mirroring_error[source] = error

    def record_serial_exported(self, source: str, serial: int) -> None:
        """
        Record an export of a source at a particular serial.
        Only the most recent serial is stored in the DB status.
        """
        self._sources_seen.add(source)
        self._exported_serials[source] = serial

    def record_operation(self, operation: DatabaseOperation, rpsl_pk: str, source: str, object_class: str,
                         object_text: str, forced_serial: Optional[int]) -> None:
        """
        Make a record in the journal of a change to an object.

        Will only record changes when self.journaling_enabled is set,
        and the database.SOURCE.keep_journal is set.
        The source will always be added to _sources_seen.

        Note that this method locks the journal table for writing to ensure a
        gapless set of NRTM serials.
        """
        self._sources_seen.add(source)
        if self.journaling_enabled and get_setting(f'sources.{source}.keep_journal'):
            journal_tablename = RPSLDatabaseJournal.__tablename__

            if forced_serial is None:
                self.database_handler.execute_statement(f'LOCK TABLE {journal_tablename} IN EXCLUSIVE MODE')
                serial_nrtm = sa.select([sa.text(f'COALESCE(MAX(serial_nrtm), 0) + 1')])
                serial_nrtm = serial_nrtm.where(RPSLDatabaseJournal.__table__.c.source == source)
                serial_nrtm = serial_nrtm.as_scalar()
            else:
                serial_nrtm = forced_serial
            stmt = RPSLDatabaseJournal.__table__.insert().values(
                rpsl_pk=rpsl_pk,
                source=source,
                operation=operation,
                object_class=object_class,
                object_text=object_text,
                serial_nrtm=serial_nrtm,
            ).returning(self.c_journal.serial_nrtm)

            insert_result = self.database_handler.execute_statement(stmt)
            inserted_serial = insert_result.fetchone()['serial_nrtm']
            self._new_serials_per_source[source].add(inserted_serial)
        elif forced_serial is not None:
            self._new_serials_per_source[source].add(forced_serial)

    def finalise_transaction(self):
        """
        - Create a new status object for all seen sources if it does not exist.
        - Reset the force_reload flag.
        - If new serials were recorded for a source, update the database
          serial stats in the status object.
        - Update the latest source errors.
        """
        for source in self._sources_seen:
            stmt = pg.insert(RPSLDatabaseStatus).values(
                source=source,
                force_reload=False,
                updated=datetime.now(timezone.utc),
            )
            stmt = stmt.on_conflict_do_update(
                index_elements=['source'],
                set_={'force_reload': False},
            )
            self.database_handler.execute_statement(stmt)

        for source, serials in self._new_serials_per_source.items():
            serial_oldest_seen = sa.select([
                sa.func.least(sa.func.min(self.c_status.serial_oldest_seen), min(serials))
            ]).where(self.c_status.source == source)
            serial_newest_seen = sa.select([
                sa.func.greatest(sa.func.max(self.c_status.serial_newest_seen), max(serials))
            ]).where(self.c_status.source == source)

            serial_oldest_journal = sa.select([
                sa.func.min(self.c_journal.serial_nrtm)
            ]).where(self.c_journal.source == source)
            serial_newest_journal = sa.select([
                sa.func.max(self.c_journal.serial_nrtm)
            ]).where(self.c_journal.source == source)

            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                serial_oldest_seen=serial_oldest_seen,
                serial_newest_seen=serial_newest_seen,
                serial_oldest_journal=serial_oldest_journal,
                serial_newest_journal=serial_newest_journal,
                updated=datetime.now(timezone.utc),
            )
            self.database_handler.execute_statement(stmt)

        for source, error in self._mirroring_error.items():
            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                last_error=error,
                last_error_timestamp=datetime.now(timezone.utc),
                updated=datetime.now(timezone.utc),
            )
            self.database_handler.execute_statement(stmt)

        for source, serial in self._exported_serials.items():
            stmt = RPSLDatabaseStatus.__table__.update().where(self.c_status.source == source).values(
                serial_last_export=serial,
            )
            self.database_handler.execute_statement(stmt)

        self._reset()

    def _reset(self):
        self._new_serials_per_source = defaultdict(set)
        self._sources_seen = set()
        self._mirroring_error = dict()
        self._exported_serials = dict()



import logging
from twisted.internet import reactor
from twisted.internet.protocol import Factory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver
from twisted.protocols.policies import TimeoutMixin

from irrd.conf import get_setting
from .query_parser import WhoisQueryParser
from .query_pipeline import QueryPipelineThread
from ..access_check import is_client_permitted

logger = logging.getLogger(__name__)


class WhoisQueryReceiver(TimeoutMixin, LineOnlyReceiver):
    """
    The query receiver is created once for each TCP connection.

    It handles interaction with the socket, and passes most work
    off to a query pipeline thread.
    """
    delimiter = b'\n'

    def connectionMade(self) -> None:  # noqa: N802
        """
        Handle a new connection. This includes
        - Access checks
        - Creating a query parser.
        - Creating a query pipeline thread and starting the thread.
        """
        peer = self.transport.getPeer()

        if not self.is_client_permitted(peer):
            self.transport.write(b'%% Access denied')
            self.transport.loseConnection()
            return

        # disable nagle
        self.transport.setTcpNoDelay(True)

        self.peer_str = f'[{peer.host}]:{peer.port}'

        self.query_parser = WhoisQueryParser(peer, self.peer_str)
        self.query_pipeline_thread = QueryPipelineThread(
            peer_str=self.peer_str,
            query_parser=self.query_parser,
            response_callback=self.query_response_callback,
            lose_connection_callback=self.lose_connection_callback,
        )
        self.query_pipeline_thread.start()

        self.setTimeout(self.query_parser.timeout)
        logger.debug(f'{self.peer_str}: new connection opened')

    def lineReceived(self, line_bytes: bytes) -> None:  # noqa: N802
        """
        Handle a line sent by a user. This method is kept as light
        as possible, to off-load most work to the pipeline thread.
        """
        self.resetTimeout()
        self.query_pipeline_thread.add_query(line_bytes)

    def query_response_callback(self, response: bytes) -> None:
        """
        Handle a response to a query, sent by the query pipeline thread.
        This is a small thread-safe wrapping around return_response()
        """
        reactor.callFromThread(self.return_response, response)

    def return_response(self, response: bytes) -> None:
        """
        Process a response. This means sending the response, closing the
        connection if needed, and telling the query pipeline thread
        sending the response completed.
        """
        self.transport.write(response)

        if not self.query_parser.multiple_command_mode:
            self.transport.loseConnection()
            logger.debug(f'{self.peer_str}: auto-closed connection')

        self.setTimeout(self.query_parser.timeout)

        self.query_pipeline_thread.ready_for_next_result()

    def lose_connection_callback(self) -> None:
        """
        Handle a !q query, as detected by the query pipeline thread.
        This is a small thread-safe wrapping around transport.loseConnection()
        """
        reactor.callFromThread(self.transport.loseConnection)

    def connectionLost(self, reason=connectionDone) -> None:  # noqa: N802
        """
        Handle a lost connection, either because transport.loseConnection()
        was called, or the remote side closed it.
        Cancels the pipeline thread, causing it to terminate within two
        seconds.
        """
        self.factory.current_connections -= 1
        self.query_pipeline_thread.cancel()

    def timeoutConnection(self) -> None:  # noqa: N802
        """
        Triggered when the configured timeout occurs. Ignored if queries are
        still running or queued, otherwise the connection is closed.
        """
        if not self.query_pipeline_thread.is_processing_queries():
            self.transport.loseConnection()

    def is_client_permitted(self, peer) -> bool:
        """
        Check whether a client is permitted.
        """
        return is_client_permitted(peer, 'server.whois.access_list', default_deny=False)


class WhoisQueryReceiverFactory(Factory):
    protocol = WhoisQueryReceiver

    def __init__(self):
        self.current_connections = 0
        self.max_connections = get_setting('server.whois.max_connections')

    def buildProtocol(self, addr):  # noqa: N802
        if self.current_connections >= self.max_connections:
            return None
        self.current_connections += 1
        protocol = self.protocol()
        protocol.factory = self
        return protocol



#!/usr/bin/env python
# flake8: noqa: E402

"""
A simple load tester for IRRd.
Sends random !g queries.
"""
import time

import argparse
import random
import socket


def main(host, port, count):
    queries = [b'!!\n']
    for i in range(count):
        asn = random.randrange(1, 50000)
        query = f'!gAS{asn}\n'.encode('ascii')
        queries.append(query)
    queries.append(b'!q\n')

    s = socket.socket()
    s.settimeout(3000)
    s.connect((host, port))

    queries = b"""!!
!aAS-CWASIA
!aAS-CWCI
!aAS-CWCS-PS
!aAS-CYBERLYNK
!aAS-CYPRESSTEL
!aAS-CYSO
!aAS-Comporium-Transit
!aAS-Cybercon
!aAS-D-NET
!aAS-DACOM
!aAS-DAILYMOTION
!aAS-DALENYS
!aAS-DATA393-INV01
!aAS-DATAHOP
!aAS-DATAPIPE
!aAS-DCI
!aAS-DCLUX
!aAS-DDOS-GUARD
!aAS-DEFENSE
!aAS-DELTA
!aAS-DENIC
!aAS-DEVCLIC
!aAS-DGINET
!aAS-DIALTELECOM
!aAS-DIDICHUXING
!aAS-DIMENOC
!aAS-DINAMICA-BR
!aAS-DIRECT_PEERS
!aAS-DIYIXIAN
!aAS-DLIVE
!aAS-DMITINC
!aAS-DOCLER
!aAS-DOTRO
!aAS-DRUKNET-TRANSIT
!aAS-DSTORAGE
!aAS-DSTORAGE-V6
!aAS-DTN
!aAS-DUMARCA
!aAS-DUOCAST
!aAS-DWNI
!aAS-DXNET
!aAS-DXTLTECH
!aAS-DYNINC
!aAS-EASYNET
!aAS-EBT
!aAS-EC-MI
!aAS-EDGECAST
!aAS-EDION
!aAS-EDISON-CARRIER-SOLUTIONS
!aAS-EDXNETWORK
!aAS-EFAR
!aAS-EHIME-CATV
!aAS-EIGI
!aAS-EIGI-SC
!aAS-EIRCOM
!aAS-EISCO
!aAS-EISON-LTD
!aAS-EITC-DU
!aAS-EL-K-STIL
!aAS-ELASTX
!aAS-ELI
!aAS-ELIAS
!aAS-ELLIJAY
!aAS-EMAILVISION
!aAS-EMIX
!aAS-ENTELBOLIVIA
!aAS-ENTER
!aAS-ENTER-V6
!aAS-EONIX
!aAS-EQUANT-ASIA
!aAS-EQUINIX-AP
!aAS-EQUINIX-EC
!aAS-EQUINIX-EU
!aAS-EQUINIX-OS
!aAS-ESH
!aAS-ESNET
!aAS-ETHEROUTE
!aAS-ETOP
!aAS-EUNETWORKS
!aAS-EUTELSAT-TH2
!aAS-EVERESTDATA
!aAS-EXA
!aAS-EXASCALE
!aAS-EXATEL-INT
!aAS-EXN
!aAS-EXPEDIENT
!aAS-EXPONENTIAL-e
!aAS-EXPONENTIAL-e-6
!aAS-EnergyGroupNetworks
!aAS-Expedient6
!aAS-FARSIGHT
!aAS-FASTNET
!aAS-FBDC
!aAS-FHE3NET
!aAS-FHL
!aAS-FIBERHUB
!aAS-FIBERRING
!aAS-FIDONET
!aAS-FIRSTCOLO
!aAS-FLAGT
!aAS-FLUENCY
!aAS-FOBUL
!aAS-FOBUL-V6
!aAS-FORCEPOINT-CLOUD
!aAS-FOREMOST
!aAS-FORTEX
!aAS-FORTEX6
!aAS-FPP
!aAS-FPT
!aAS-FREECOMM
!aAS-FREENETDE
!aAS-FREENETDE6
!aAS-FROM-JP-FTMNET
!aAS-FSYS
!aAS-FUSE
!aAS-FUSIX
!aAS-FUZENET
!aAS-G8
!aAS-GAMMATELECOM
!aAS-GBLX
!aAS-GBXS
!aAS-GBXS6
!aAS-GCI
!aAS-GCNET
!aAS-GCNETV6
!aAS-GCOM-MD
!aAS-GCONNECT
!aAS-GCORE
!aAS-GCTR-JP
!aAS-GDG
!aAS-GENCAT
!aAS-GENESISADAPTIVE
!aAS-GFSB
!aAS-GIGAINFRA
!aAS-GIGAINFRA-V6
!aAS-GIGANEWS-ALL
!aAS-GIPNET
!aAS-GITS
!aAS-GLIDE
!aAS-GLOBAL
!aAS-GLOBALNOC
!aAS-GLOBALNOC6
!aAS-GLOBEINTERNET
!aAS-GLOBEINTERNET-IPv6
!aAS-GMO
!aAS-GNT
!aAS-GOOGLE
!aAS-GORILLASERVERS
!aAS-GOSCOMB
!aAS-GOSCOMB-V6
!aAS-GRAFIX-TRANSIT
!aAS-GRAJAUNET
!aAS-GRU
!aAS-GSVNET
!aAS-GTLD
!aAS-GVA
!aAS-GVM
!aAS-GVM6
!aAS-GXT
!aAS-GYRON
!aAS-GigaDefence
!aAS-HANABI
!aAS-HANABI4-MDDOS
!aAS-HANOITELECOM
!aAS-HARGRAY-TRANSIT
!aAS-HEAS
!aAS-HELINET-V4
!aAS-HELINET-V6
!aAS-HETZNER
!aAS-HGC-INTL
!aAS-HIBERNIA
!aAS-HIGE
!aAS-HIGHWINDS
!aAS-HKBN
!aAS-HKNET
!aAS-HOPONE
!aAS-HOSTIT-MK
!aAS-HOSTIT-NN
!aAS-HOSTMY
!aAS-HOSTSERVER
!aAS-HOSTWAY
!aAS-HOTnet-HOTCN
!aAS-HPC-MVM
!aAS-HURRICANE
!aAS-HURRICANEv6
!aAS-HUTCHCITY
!aAS-I-NET
!aAS-I123
!aAS-I3-BGP-CUSTOMERS
!aAS-IBGC
!aAS-IBMNET-JP
!aAS-IBOSS
!aAS-ICANN
!aAS-ICANN6
!aAS-ICNC
!aAS-ICTX
!aAS-IDAQ
!aAS-IDC
!aAS-IDC6
!aAS-IDNET
!aAS-IGNEMEA
!aAS-IGUANE
!aAS-IGW-SET
!aAS-IHENDERSON
!aAS-II-OKINAWA
!aAS-IIJ
!aAS-IIJ6
!aAS-IMGIX
!aAS-IN2NET
!aAS-INCAPSULA
!aAS-INCOMM
!aAS-INETBONE
!aAS-INETBONE-V6
!aAS-INETC
!aAS-INFOPACT
!q""".splitlines()

    queries_str = b'\n'.join(queries) + b'\n'
    s.sendall(queries_str)

    start_time = time.perf_counter()
    while 1:
        data = s.recv(1024*1024)
        if not data:
            break

    count = len(queries) - 1
    elapsed = time.perf_counter() - start_time
    time_per_query = elapsed / count * 1000
    qps = int(count / elapsed)
    print(f'Ran {count} queries in {elapsed}s, time per query {time_per_query} ms, {qps} qps')


if __name__ == '__main__':  # pragma: no cover
    description = """A simple load tester for IRRd. Sends random !g queries."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--count', dest='count', type=int, default=5000,
                        help=f'number of queries to run (default: 5000)')
    parser.add_argument('host', type=str,
                        help='hostname of instance')
    parser.add_argument('port', type=int,
                        help='port of instance')
    args = parser.parse_args()

    main(args.host, args.port, args.count)



import time

import logging
import queue
import threading
from typing import Callable

from .query_parser import WhoisQueryParser

logger = logging.getLogger(__name__)


class QueryPipelineThread(threading.Thread):
    """
    The query pipeline thread is a thread in charge of running queries.

    IRRd queries need to be run sequentially, and support a large amount of
    queries being sent at once, which then need be answered one by one,
    in the order they were sent.
    """
    def __init__(self, peer_str: str, query_parser: WhoisQueryParser,
                 response_callback: Callable[[bytes], None], lose_connection_callback: Callable[[], None],
                 *args, **kwargs):
        """
        :param peer_str: a string representing the connection's peer
        :param query_parser: an instance of WhoisQueryParser
        :param response_callback: a method to call with a bytes argument,
               when a response must be sent
        :param lose_connection_callback: a method to call when the connection
               should be closed
        """
        self.peer_str = peer_str
        self.query_parser = query_parser
        self.response_callback = response_callback
        self.lose_connection_callback = lose_connection_callback

        self.pipeline: queue.Queue[bytes] = queue.Queue()
        self.ready_to_send_result = threading.Event()
        self.ready_to_send_result.set()
        self.cancelled = False
        self.processing_query = False

        super().__init__(*args, **kwargs)

    def add_query(self, query: bytes) -> None:
        """
        Add a query to the pipeline to be processed.
        Thread safe. Typically called from twisted's main thread,
        when a query is received.
        """
        self.pipeline.put(query, block=False)

    def is_processing_queries(self) -> bool:
        """
        Check whether this thread is processing a query and/or still has
        queries in the pipeline.
        """
        return self.processing_query or not self.pipeline.empty()

    def cancel(self) -> None:
        """
        Schedule this thread for cancellation, possibly with a delay.
        Thread safe. Typically called from twisted's main thread,
        when the connection is closed for whatever reason.
        (Threads can't otherwise be killed easily in Python.)
        """
        self.cancelled = True

    def run(self) -> None:  # pragma: no cover
        """
        Main thread running method. Attempts to process newly added queries
        one by one, until self.cancelled is set.
        """
        while not self.cancelled:
            self._fetch_process_query()

    def _fetch_process_query(self) -> None:
        """
        Attempt to fetch and process a query.

        Returns either when a query was processed, or every 2 seconds if no
        queries are in the queue. This allows the outer loop in run() to check
        whether this thread should be cancelled due to a closed connection.

        See also ready_for_next_result().
        """
        try:
            query_bytes = self.pipeline.get(block=True, timeout=2)
        except queue.Empty:
            return

        self.processing_query = True
        start_time = time.perf_counter()
        query = query_bytes.decode('utf-8', errors='backslashreplace').strip()

        if not query:
            self.processing_query = False
            return

        logger.info(f'{self.peer_str}: processing query: {query}')

        if query.upper() == '!Q':
            self.lose_connection_callback()
            logger.debug(f'{self.peer_str}: closed connection per request')
            self.cancel()
            self.processing_query = False
            return

        response = self.query_parser.handle_query(query)
        response_bytes = response.generate_response().encode('utf-8')

        self.ready_to_send_result.wait()
        self.ready_to_send_result.clear()
        self.response_callback(response_bytes)

        elapsed = time.perf_counter() - start_time
        logger.info(f'{self.peer_str}: sent answer to query, elapsed {elapsed}s, {len(response_bytes)} bytes: {query}')
        self.processing_query = False

    def ready_for_next_result(self) -> None:
        """
        Reset the lock, allowing a new call to response_callback.

        To not overwhelm the Twisted main thread, the response_callback will
        only be called once at first. After that, the lock needs to be cleared
        by calling this method.
        """
        self.ready_to_send_result.set()



import sqlalchemy as sa
import ujson

from irrd.conf import get_setting

engine = None


def get_engine():
    global engine
    if not engine:
        database_pool_size = (get_setting('server.whois.max_connections') + len(get_setting('sources', []))) * 3
        engine = sa.create_engine(
            get_setting('database_url'),
            pool_size=database_pool_size,
            max_overflow=30,
            json_deserializer=ujson.loads,
        )

    return engine



# flake8: noqa: E402
import sys

import os
from pathlib import Path
from twisted.application import internet, service
from twisted.application.service import IServiceMaker
from twisted.plugin import IPlugin
from twisted.python import usage
from twisted.python.log import PythonLoggingObserver, ILogObserver
from zope.interface import implementer

sys.path.append(str(Path(__file__).resolve().parents[1]))

from irrd.conf import config_init, get_setting, CONFIG_PATH_DEFAULT
from irrd.mirroring.scheduler import MirrorScheduler
from irrd.server.http.http_resources import http_site
from irrd.server.whois.protocol import WhoisQueryReceiverFactory


class Options(usage.Options):
    optParameters = [['config', 'c', CONFIG_PATH_DEFAULT, 'Path to the IRRd config file.']]


class PythonLoggingMultiService(service.MultiService):
    def setServiceParent(self, parent):
        service.MultiService.setServiceParent(self, parent)
        parent.setComponent(ILogObserver, PythonLoggingObserver().emit)


@implementer(IServiceMaker, IPlugin)
class ServiceMaker(object):
    tapname = 'irrd'
    description = 'Internet Routing Registry daemon'
    options = Options

    def makeService(self, options):
        config_init(options['config'])

        ms = PythonLoggingMultiService()

        internet.TCPServer(
            get_setting('server.whois.port'),
            WhoisQueryReceiverFactory(),
            interface=get_setting('server.whois.interface')
        ).setServiceParent(ms)

        internet.TCPServer(
            get_setting('server.http.port'),
            http_site,
            interface=get_setting('server.http.interface')
        ).setServiceParent(ms)

        mirror_frequency = int(os.environ.get('IRRD_SCHEDULER_TIMER_OVERRIDE', 15))
        internet.TimerService(mirror_frequency, MirrorScheduler().run).setServiceParent(ms)

        return ms


serviceMaker = ServiceMaker()



import itertools
import math
from collections import defaultdict

import logging
import os
import signal
import threading
from typing import Optional, List, Set, Dict

from .queries import RPSLDatabaseQuery

RELOAD_SIGNAL = signal.SIGUSR1

_preloader = None

logger = logging.getLogger(__name__)


class Preloader:
    """
    The preloader allows information to be preloaded into memory.

    For queries that repeat often, or repeatedly retrieve nearly the same,
    large data sets, this can improve performance.

    The DatabaseHandler calls reload() when data has been changed, added or
    deleted, which may then cause a thread to be scheduled to update the
    store, or conclude an update is already pending anyways.
    """
    def __init__(self):
        self._origin_route4_store = defaultdict(set)
        self._origin_route6_store = defaultdict(set)

        self._reload_lock = threading.Lock()
        self._threads = []

        self._store_ready_event = threading.Event()
        self.reload()

    def routes_for_origins(self, origins: List[str], ip_version: Optional[int]=None) -> Set[str]:
        """
        Retrieve all prefixes (in str format) originating from the provided origins.

        Prefixes are guaranteed to be unique. ip_version can be set to 4 or 6
        to restrict responses to IPv4 or IPv6 prefixes. Blocks until the first
        store has been built.
        Origins must be strings in a cleaned format, e.g. AS65537, but not
        AS065537 or as65537.
        """
        self._store_ready_event.wait()

        if ip_version and ip_version not in [4, 6]:
            raise ValueError(f'Invalid IP version: {ip_version}')

        prefix_sets: List[Set[str]] = list()
        if not ip_version or ip_version == 4:
            prefix_sets = prefix_sets + [self._origin_route4_store.get(k, set()) for k in origins]
        if not ip_version or ip_version == 6:
            prefix_sets = prefix_sets + [self._origin_route6_store.get(k, set()) for k in origins]

        return set(itertools.chain.from_iterable(prefix_sets))

    def reload(self, object_classes_changed: Optional[Set[str]]=None) -> None:
        """
        Perform a (re)load.
        Should be called after changes to the DB have been committed.

        This will start a new thread to reload the store. If a thread is
        already running, the new thread will start after the current one
        is done, due to locking.

        If a current thread is running, and a next thread is already
        running as well (waiting for a lock) no action is taken. The
        change that prompted this reload call will already be processed
        by the thread that is currently waiting.

        If object_classes_changed is provided, a reload is only performed
        if those classes are relevant to the data in the preload store.
        """
        relevant_object_classes = {'route', 'route6'}
        if object_classes_changed is not None and not object_classes_changed.intersection(relevant_object_classes):
            return
        self._remove_dead_threads()
        if len(self._threads) > 1:
            # Another thread is already scheduled to follow the current one
            return
        thread = PreloadUpdater(self, self._reload_lock, self._store_ready_event)
        thread.start()
        self._threads.append(thread)

    def update_route_store(self, new_origin_route4_store, new_origin_route6_store):
        self._origin_route4_store = new_origin_route4_store
        self._origin_route6_store = new_origin_route6_store

    def _remove_dead_threads(self) -> None:
        """
        Remove dead threads from self.threads(),
        for an accurate count of how many are alive.
        """
        for thread in self._threads:
            if not thread.is_alive():
                self._threads.remove(thread)


class PreloadUpdater(threading.Thread):
    """
    PreloadUpdater is a thread that updates the preload store,
    currently for prefixes per origin per address family.
    """
    def __init__(self, preloader, reload_lock, store_ready_event, *args, **kwargs):
        self.preloader = preloader
        self.reload_lock = reload_lock
        self.store_ready_event = store_ready_event
        super().__init__(*args, **kwargs)

    def run(self, mock_database_handler=None) -> None:
        """
        Main thread runner function.

        The reload_lock ensures only a single instance can run at the same
        time, i.e. if two threads are started, one will wait for the other
        to finish.

        After loading the data from the database, sets the two new stores
        on the provided preloader object.
        The lock is then released to allow another thread to start, and
        the store_ready_event set to indicate that the store has been
        loaded at least once, and answers can be provided based on it.

        For tests, mock_database_handler can be used to provide a mock.
        """
        self.reload_lock.acquire()
        logger.debug(f'Starting preload store update from thread {self}')

        new_origin_route4_store: Dict[str, Set] = defaultdict(set)
        new_origin_route6_store: Dict[str, Set] = defaultdict(set)

        if not mock_database_handler:  # pragma: no cover
            from .database_handler import DatabaseHandler
            dh = DatabaseHandler()
        else:
            dh = mock_database_handler

        q = RPSLDatabaseQuery(column_names=['ip_version', 'ip_first', 'ip_size', 'asn_first'], enable_ordering=True)
        q = q.object_classes(['route', 'route6'])

        for result in dh.execute_query(q):
            prefix = result['ip_first']
            key = 'AS' + str(result['asn_first'])

            if result['ip_version'] == 4:
                length = int(32 - math.log2(result['ip_size']))
                new_origin_route4_store[key].add(f'{prefix}/{length}')
            if result['ip_version'] == 6:
                length = int(128 - math.log2(result['ip_size']))
                new_origin_route6_store[key].add(f'{prefix}/{length}')

        dh.close()

        self.preloader.update_route_store(new_origin_route4_store, new_origin_route6_store)

        logger.info(f'Completed updating preload store from thread {self}, '
                    f'loaded v4 routes for {len(new_origin_route4_store)} ASes, '
                    f'v6 routes for {len(new_origin_route6_store)} ASes')
        self.reload_lock.release()
        self.store_ready_event.set()


def get_preloader():
    global _preloader
    if not _preloader:
        _preloader = Preloader()
    return _preloader


def reload_signal_handler(signum, frame):
    """
    Reload the preload store when the signal is received.
    """
    get_preloader().reload()


def send_reload_signal(irrd_pidfile):
    """
    Send a reload signal to an IRRd instance given its pidfile.

    This is used from scripts like submit_email and load_database,
    which make direct changes to the SQL database, but can't signal
    the preloader directly from their DatabaseHandler - as they
    are part of a different process.
    """
    try:
        with open(irrd_pidfile) as fh:
            irrd_pid = fh.read()
            try:
                os.kill(int(irrd_pid), RELOAD_SIGNAL)
            except (ProcessLookupError, ValueError):
                logger.warning(f'Attempted to send reload signal to update preloader for IRRD on '
                               f'PID {irrd_pid}, but process is not running or PID is invalid.')
    except OSError as ose:
        logger.warning(f'Attempted to send reload signal to update preloader for IRRD with PID '
                       f'from {irrd_pidfile}, but file could not be opened: {ose}')


signal.signal(RELOAD_SIGNAL, reload_signal_handler)



#!/usr/bin/env python
# flake8: noqa: E402

"""
A simple load tester for IRRd.
Sends random !g queries.
"""
import time

import argparse
import random
import socket


def main(host, port, count):
    queries = [b'!!\n']
    for i in range(count):
        asn = random.randrange(1, 50000)
        query = f'!gAS{asn}\n'.encode('ascii')
        queries.append(query)
    queries.append(b'!q\n')

    s = socket.socket()
    s.settimeout(600)
    s.connect((host, port))

    queries = b"""!!
!aAS-RAGINGWIRE
!aAS-RAINBOW
!aAS-RAKUTEN-RBN
!aAS-RATIONAL-CUS
!aAS-RBLHST
!aAS-RDSNET
!aAS-REBACOM
!aAS-REDIRIS
!aAS-RETELIT
!aAS-RETN
!aAS-RETN6
!aAS-RG-DFW
!aAS-RG-IAD
!aAS-RG-SEA
!aAS-RHTEC
!aAS-RHTEC6
!aAS-RICHTOPEC
!aAS-RINGCENTRAL
!aAS-RIOT
!aAS-RKOM
!aAS-RMI
!aAS-RMIFL
!aAS-RNP-TRANSIT
!aAS-ROCKISLAND
!aAS-ROKA
!aAS-ROSITE
!aAS-ROSITE6
!aAS-RPS
!aAS-RTD
!aAS-RUBICONPROJECT
!aAS-RWTS-DS
!aAS-SAKURA
!aAS-SALESFORCE
!aAS-SANNET
!aAS-SARENET
!aAS-SARGASSO-UK
!aAS-SAVVISJP
!aAS-SBN-IIG
!aAS-SCIS
!aAS-SDT
!aAS-SDT-V6
!aAS-SEABONE
!aAS-SEABONE-V6
!aAS-SECARMA
!aAS-SECBUBWC
!aAS-SECH-EAT
!aAS-SECURA
!aAS-SEEDNET
!aAS-SEMAPHORE
!aAS-SERVERCENTRAL
!aAS-SERVEREL
!aAS-SERVERSCOM
!aAS-SET-40356
!aAS-SET-9930
!aAS-SET-AFRIX
!aAS-SET-ANTAMEDIAKOM
!aAS-SET-BIGBAND
!aAS-SET-BSCCL
!aAS-SET-GTI
!aAS-SET-IPDC
!aAS-SET-SEACOM
!aAS-SET-TELBRU-APNIC
!aAS-SET-XLCG-GLOBAL
!aAS-SET6-38566
!aAS-SETPTCL
!aAS-SGGS
!aAS-SHARKS
!aAS-SHARKS6
!aAS-SHENTEL
!aAS-SINET4
!aAS-SINGTEL
!aAS-SIOL
!aAS-SIOL6
!aAS-SIP
!aAS-SKBroadband
!aAS-SKYCABLE
!aAS-SKYVISION
!aAS-SLF-13760-DOWNSTREAM
!aAS-SLINC
!aAS-SLT-GLOBAL
!aAS-SMAID
!aAS-SO-NET6
!aAS-SOFIA-CONNECT
!aAS-SOFIA-CONNECT-V6
!aAS-SOFTLAYER
!aAS-SOHONET-UK
!aAS-SOLCON
!aAS-SOLCON6
!aAS-SOLNET
!aAS-SOMEADDRESS
!aAS-SONY-CGEI
!aAS-SOTACONNECT
!aAS-SOUTHEAST-TEXAS-GIGAPOP-VERIO
!aAS-SOVAM
!aAS-SPACENET
!aAS-SPACENET-V6
!aAS-SPACENETWORK
!aAS-SPNET
!aAS-SPOTIFY
!aAS-SRT
!aAS-SSCNETWORKS
!aAS-STACKPATH
!aAS-STARHUBINTERNET
!aAS-STARTNIX-EU
!aAS-STARTOUCH
!aAS-STCN
!aAS-STEADFAST
!aAS-STEALTH
!aAS-STRATUS-DOWNSTREAM
!aAS-STS-RO
!aAS-SUITABLETECH
!aAS-SUPERINTERNET
!aAS-SUPRANET
!aAS-SURETEC
!aAS-SURFNET
!aAS-SWCMGLOBAL
!aAS-SWIPNET
!aAS-SWITCHCO
!aAS-SYMC
!aAS-SYSELEVEN
!aAS-SZARANET
!aAS-So-Net
!aAS-TABOOLA
!aAS-TANGO
!aAS-TASHICELL
!aAS-TCTWest
!aAS-TDATANET
!aAS-TDCNET
!aAS-TDCNET-IPV6
!aAS-TDDE
!aAS-TECHNO
!aAS-TECHNO-V6
!aAS-TECNOCRATICA
!aAS-TELAPPLIANT
!aAS-TELCONET-TRANSIT-AS
!aAS-TELE2
!aAS-TELENET
!aAS-TELEPAK-C
!aAS-TELETRANS
!aAS-TELIKO
!aAS-TELMEX
!aAS-TELSTRA-INTL-CUSTOMERS
!aAS-TERTIARY
!aAS-TFN-TRANSIT1
!aAS-TGW
!aAS-THEGIGABIT
!aAS-THINFACTORY
!aAS-THIX
!aAS-THRESHINC
!aAS-TICGW
!aAS-TIG-GROUP
!aAS-TIGGEE
!aAS-TIMICO
!aAS-TIPNETIT
!aAS-TKPSA-UP
!aAS-TMHK
!aAS-TMNET-CUSTOMERS
!aAS-TMR
!aAS-TMX-PERU-TRANSIT
!aAS-TOTALUPTIME
!aAS-TOUCHTONE
!aAS-TOWARDEX
!aAS-TRANSIP
!aAS-TRANSIT-GLOBE
!aAS-TRANSITOS
!aAS-TRANSIX-E
!aAS-TRANSIX-W
!aAS-TREML-STURM
!aAS-TRIPLETNET
!aAS-TROYC-01
!aAS-TRUE
!aAS-TRUNKNETWORKS
!aAS-TTK
!aAS-TTK6
!aAS-TTNET
!aAS-TURN
!aAS-TVC
!aAS-TWIG
!aAS-TWITTER
!aAS-TWTIT
!aAS-Tecnocratica
!aAS-Telx
!aAS-Transit-NetNam
!aAS-U-NETSURF
!aAS-UCC
!aAS-UKBB6
!aAS-UKSERVERS
!aAS-ULTRACOM
!aAS-UNBELIEVABLE
!aAS-UNILINK
!aAS-UNILINK-V6
!aAS-UNILOGICNET
!aAS-UNISERVER
!aAS-UNITAS
!aAS-VAIONI
!aAS-VALVE
!aAS-VDC
!aAS-VECTANT
!aAS-VELIANET
!aAS-VELIANET-V6
!aAS-VERIO-65029
!aAS-VERIO-65031
!aAS-VERIO-65052
!aAS-VERIO-65053
!aAS-VERIO-65056
!aAS-VERIO-65061
!aAS-VERIO-65072
!aAS-VERIO-65073
!aAS-VERIO-65301
!aAS-VERIO-65304
!aAS-VERIXI
!aAS-VERSATEL-V6
!aAS-VIANET
!aAS-VIAWEST-TRANSIT
!aAS-VIPNETNTT
!aAS-VIRTELA-COMM
!aAS-VKONTAKTE
!aAS-VORBOSS
!aAS-VORBOSS6
!aAS-VOXILITY-SET
!aAS-VRNETZE-FRA
!aAS-WBTSJP-NET
!aAS-WEBAIR
!aAS-WEBDISCOUNT
!aAS-WEBEX
!aAS-WEBNX
!aAS-WERADIO
!aAS-WERITECH
!aAS-WIFIRST
!aAS-WIKIMEDIA
!aAS-WILDCARD
!aAS-WIMANX
!aAS-WINDSTREAM
!aAS-WINKY
!aAS-WINSPEED-TRANSIT
!aAS-WOLCOMM
!aAS-WOW
!aAS-WPT
!aAS-WWN
!aAS-WZ-EU
!aAS-WZ-US
!aAS-XEPHION
!aAS-XFERNET
!aAS-XMISSION
!aAS-XO
!aAS-XPNTI
!aAS-XTOM
!aAS-XTRAORDINARY
!aAS-ZAYO-ZCLOUD-ATL01
!aAS-ZEN
!aAS-ZENEDGE
!aAS-ZENLAYER
!aAS-ZILLO1
!aAS-anch-global
!aAS-enjoyvc-Kwai
!aAS-eo
!aAS-fnetlink
!aAS-globalfrag
!aAS-iTSCOM
!aAs-simwood
!aas-30666
!aas-AS19905
!aas-NewEdge
!aas-Peak10
!aas-cdn77
!aas-comms365
!aas-ebaymtbb
!aas-freethought
!aas-gorannet
!aas-gtld
!aas-gtt
!aas-infinitytelecom
!aas-iwonco-default
!aas-pldt
!aas-portlane
!aas-roblox
!aas-set-ipserverone
!q""".splitlines()

    queries_str = b'\n'.join(queries) + b'\n'
    s.sendall(queries_str)

    start_time = time.perf_counter()
    while 1:
        data = s.recv(1024*1024)
        if not data:
            break

    count = len(queries) - 1
    elapsed = time.perf_counter() - start_time
    time_per_query = elapsed / count * 1000
    qps = int(count / elapsed)
    print(f'Ran {count} queries in {elapsed}s, time per query {time_per_query} ms, {qps} qps')


if __name__ == '__main__':  # pragma: no cover
    description = """A simple load tester for IRRd. Sends random !g queries."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--count', dest='count', type=int, default=5000,
                        help=f'number of queries to run (default: 5000)')
    parser.add_argument('host', type=str,
                        help='hostname of instance')
    parser.add_argument('port', type=int,
                        help='port of instance')
    args = parser.parse_args()

    main(args.host, args.port, args.count)



import time
from collections import defaultdict

import logging
import threading
from typing import Dict

from irrd.conf import get_setting
from irrd.conf.defaults import DEFAULT_SOURCE_IMPORT_TIMER, DEFAULT_SOURCE_EXPORT_TIMER
from .mirror_runners_export import SourceExportRunner
from .mirror_runners_import import MirrorImportUpdateRunner

logger = logging.getLogger(__name__)


class MirrorScheduler:
    """
    Scheduler for mirroring processes.

    For each time run() is called, will start a thread for each mirror database
    unless a thread is still running for that database (which is likely to be
    the case in some full imports).
    """
    threads: Dict[str, threading.Thread]
    last_started_time: Dict[str, int]

    def __init__(self):
        self.threads = dict()
        self.last_started_time = defaultdict(lambda: 0)

    def run(self) -> None:
        for source in get_setting('sources').keys():
            is_mirror = get_setting(f'sources.{source}.import_source') or get_setting(f'sources.{source}.nrtm_host')
            import_timer = int(get_setting(f'sources.{source}.import_timer', DEFAULT_SOURCE_IMPORT_TIMER))

            if is_mirror:
                self.run_if_relevant(source, MirrorImportUpdateRunner, import_timer)

            runs_export = get_setting(f'sources.{source}.export_destination')
            export_timer = int(get_setting(f'sources.{source}.export_timer', DEFAULT_SOURCE_EXPORT_TIMER))

            if runs_export:
                self.run_if_relevant(source, SourceExportRunner, export_timer)

    def run_if_relevant(self, source: str, runner_class, timer: int):
        thread_name = f'Thread-{runner_class.__name__}-{source}'

        current_time = time.time()
        has_expired = (self.last_started_time[thread_name] + timer) < current_time
        if not has_expired or self._is_thread_running(thread_name):
            return

        logger.debug(f'Started new thread {thread_name} for mirror import/export for {source}')
        initiator = runner_class(source=source)
        thread = threading.Thread(target=initiator.run, name=thread_name)
        self.threads[thread_name] = thread
        thread.start()
        self.last_started_time[thread_name] = int(current_time)

    def _is_thread_running(self, thread_name: str):
        if thread_name not in self.threads:
            return False
        return self.threads[thread_name].is_alive()



def flatten_mock_calls(mock):
    """
    Flatten the calls performed on a particular mock object,
    into a list of calls with arguments.
    """
    result = []
    for call in mock.mock_calls:
        call = list(call)
        call_name = call[0]
        if '.' in str(call_name):
            call_name = str(call_name).split('.')[-1]
        result.append([call_name] + call[1:])
    return result



#!/usr/bin/env python
# flake8: noqa: E402

"""
A simple load tester for IRRd.
Sends random !g queries.
"""
import time

import argparse
import random
import socket


def main(host, port, count):
    queries = [b'!!\n']
    for i in range(count):
        asn = random.randrange(1, 50000)
        query = f'!gAS{asn}\n'.encode('ascii')
        queries.append(query)
    queries.append(b'!q\n')

    s = socket.socket()
    s.settimeout(600)
    s.connect((host, port))

    queries = b"""!!
!aAS-PTISP
!aAS-QRATOR
!aAS-QTEL-SET
!aAS-QTS-11767
!aAS-QTS-20141
!aAS-QTS-394968
!aAS-QTS-40107
!aAS-QTS-40913
!aAS-QTS-4136
!aAS-QTS-53907
!aAS-QWEST
!aAS-RACE
!aAS-RACKCORP
!aAS-RACKSENSE
!aAS-RACKSENSE-IPV6
!aAS-RACKSPACE
!aAS-PHONOSCOPE
!aAS-PHYBER
!aAS-PLAYONLINE
!aAS-PLIXDC-MIX
!aAS-PLUSLINE
!aAS-PNAPACS
!aAS-PNAPAMS004
!aAS-PNAPCHG
!aAS-PNAPDALZ
!aAS-PNAPFRA002
!aAS-PNAPHKG001
!aAS-PNAPIRLAB
!aAS-PNAPLAX
!aAS-PNAPLON
!aAS-PNAPMIA
!aAS-PNAPNYM
!aAS-PNAPOSK001
!aAS-PNAPPAR
!aAS-PNAPSEF
!aAS-PNAPSJE
!aAS-PNAPSYD001
!aAS-PNAPTOK
!aAS-PNAPWDC002
!aAS-PNW-GIGAPOP-I
!aAS-POBOX
!aAS-POCKETINET
!aAS-PRIMACOM-ISP
!aAS-PRIMEXM-US
!aAS-PRLSS
!aAS-PROBENETWORKS
!aAS-PROBENETWORKS-v6
!aAS-PROLE
!aAS-PROLOCATION
!aAS-PROLOCATION-V6
!aAS-PROPER
!aAS-PROPER6
!aAS-PROTEXIA
!aAS-PSINET-US
!aAS-PSLIGHTWAVE
!aAS-PSN-ID
!aAS-PTA
!aAS-INFORELAY
!aAS-INFOSPHERE
!aAS-INFOWEB
!aAS-INFOWEB-V6
!aAS-INIT7
!aAS-INNET
!aAS-INOC
!aAS-INSTART
!aAS-INTERACTIVE3D
!aAS-INTERDOMINIOS
!aAS-INTERNETONE_CH
!aAS-INTEROUTE
!aAS-INTERVIA
!aAS-IOKO
!aAS-IOVATION
!aAS-IP-MAX
!aAS-IP-MAX-V6
!aAS-IPACCT
!aAS-IPB
!aAS-IPIFONY
!aAS-IPTRANSIT
!aAS-IQNETWORKS
!aAS-IRON
!aAS-IRONDNS
!aAS-IRONDNS-V6
!aAS-IS
!aAS-ISPNET
!aAS-ITANDTEL
!aAS-ITANDTEL-IPV6
!aAS-ITHOLDINGSCUSTOMERS
!aAS-ITSTG
!aAS-IUSA
!aAS-IVANET-ASIA
!aAS-IVZ
!aAS-IXREACH
!aAS-JAGUAR
!aAS-JAGUAR-V6
!aAS-JANIS
!aAS-JOINTTRANSIT
!aAS-JPRS
!aAS-JUMP
!aAS-KAKAO
!aAS-KCOMSPN
!aAS-KDDI6
!aAS-KDDIDMPEER4
!aAS-KEYCOM
!aAS-KINGSTONUK
!aAS-KINX-TRANSIT
!aAS-KMS
!aAS-KOMTEL
!aAS-KORNET
!aAS-KPNv6
!aAS-KPTNETWORK
!aAS-KQ
!aAS-KRYPTON
!aAS-KUENTOS-TRANSIT
!aAS-LAGIS
!aAS-LAGIS6
!aAS-LAHAI-SG
!aAS-LARENF
!aAS-LATINIP
!aAS-LCHOST
!aAS-LEASEWEB
!aAS-LEASEWEB-APAC
!aAS-LEASEWEB-NL
!aAS-LIBERTYI
!aAS-LINKAGE
!aAS-LINKEDIN-TRANSIT
!aAS-LINODE
!aAS-LINXTELECOM
!aAS-LIQUID
!aAS-LONSDALESOLWAY
!aAS-LOSNETTOS
!aAS-LOSNETTOS-IPv6
!aAS-LOTTOMATICA
!aAS-LROOT
!aAS-LSNET
!aAS-LUME
!aAS-LUMOS
!aAS-LUXNETWORK
!aAS-LWL
!aAS-LeaseNet-Customers
!aAS-MACROLYNX
!aAS-MACTELECOM
!aAS-MAEHDROS
!aAS-MANDA
!aAS-MCNET
!aAS-MEDIASAT
!aAS-MEDIAWEB-NET
!aAS-MEGADIAL
!aAS-MEGASPACE
!aAS-MELBICOM
!aAS-MEON
!aAS-MERULA
!aAS-MESH
!aAS-METROCAST
!aAS-METRONETUK
!aAS-METROSWITCH
!aAS-MEX
!aAS-MEX6
!aAS-MF-MGSM
!aAS-MFE-TRANSIT
!aAS-MHDD
!aAS-MICHNET
!aAS-MICROSOFT
!aAS-MIDCO
!aAS-MISTRAL
!aAS-MITENE
!aAS-MK-NETZDIENSTE
!aAS-MK-NETZDIENSTE-V6
!aAS-MNETDE
!aAS-MNS
!aAS-MOBILY
!aAS-MOHAWK-TRANSIT
!aAS-MOJI
!aAS-MTCCOMM
!aAS-MTEL-MO
!aAS-MYNET
!aAS-MYTHIC
!aAS-NBIP
!aAS-NBIP6
!aAS-NCOM
!aAS-NCSC-DNS
!aAS-NET1OUT
!aAS-NET2ONE
!aAS-NETACTUATE
!aAS-NETASSIST
!aAS-NETCONNEX
!aAS-NETERRA-INT
!aAS-NETFORTRIS
!aAS-NETGUARDBG
!aAS-NETHER
!aAS-NETIRD
!aAS-NETIRD-V6
!aAS-NETLOGICS
!aAS-NETNOD
!aAS-NETRIVER
!aAS-NETROUTING
!aAS-NETROUTING6
!aAS-NETUSE
!aAS-NETVISIONTEL
!aAS-NETWISE
!aAS-NETWORKPLATFORMS
!aAS-NETWORXBG
!aAS-NEWMEDIAEXPRESS
!aAS-NEXGEN
!aAS-NEXICA
!aAS-NEXICOM
!aAS-NEXTGLOBALSERV
!aAS-NEXTGLOBALSERV6
!aAS-NEXTLAYER
!aAS-NEXTLAYER-V6
!aAS-NEXUS-TELECOM
!aAS-NFCR
!aAS-NFL
!aAS-NFORCE
!aAS-NHN
!aAS-NHN-TECHORUS
!aAS-NIANDC
!aAS-NIANET
!aAS-NIANETV6
!aAS-NITEL-TRANSIT
!aAS-NOANET-PEERING
!aAS-NODE4
!aAS-NOELC
!aAS-NOMICAL
!aAS-NOMINET-NS
!aAS-NOMINET-RES1
!aAS-NOMINUM-SKYE
!aAS-NOOR
!aAS-NORDUNET
!aAS-NORIS
!aAS-NOVATEL
!aAS-NSCS-NET
!aAS-NSI-DC
!aAS-NSL-CUSTOMERS
!aAS-NSONE
!aAS-NT
!aAS-NTT-CISCO
!aAS-NTTCOMWARE
!aAS-NTTNETID
!aAS-NULL
!aAS-NUWAVE
!aAS-NYNEX
!aAS-Novatectelecom
!aAS-OBENETWORK
!aAS-OC-ASSET
!aAS-OCG
!aAS-OCN-ALL
!aAS-OCN-ALL-V6
!aAS-ODN
!aAS-ODN-V6
!aAS-ONERINGNETWORKS
!aAS-ONO-IPV6
!aAS-ONO-UPS
!aAS-ONYX
!aAS-OPAL
!aAS-OPENCARRIER
!aAS-OPENCARRIER6
!aAS-OPENDNS
!aAS-OPENTRANSIT
!aAS-OPERA
!aAS-OPERA-ASIA
!aAS-OPTINET
!aAS-ORBITAL
!aAS-OTL
!aAS-OVH
!aAS-OZ-ITHS
!aAS-PACIFICNET
!aAS-PACNET-DOWNSTREAM
!aAS-PALTEL
!aAS-PANTEL
!aAS-PARKER
!aAS-PARTAN
!aAS-PAVLOV
!aAS-PAXIO
!aAS-PAYPAL
!aAS-PCH
!aAS-PEGTECHCUSTOMERS
!aAS-PENGUIN
!aAS-PEOPLESCOM
!aAS-PHIBEE-TELECOM
!aAS-PHOENIXNAP
!q""".splitlines()

    queries_str = b'\n'.join(queries) + b'\n'
    s.sendall(queries_str)

    start_time = time.perf_counter()
    while 1:
        data = s.recv(1024*1024)
        if not data:
            break

    count = len(queries) - 1
    elapsed = time.perf_counter() - start_time
    time_per_query = elapsed / count * 1000
    qps = int(count / elapsed)
    print(f'Ran {count} queries in {elapsed}s, time per query {time_per_query} ms, {qps} qps')


if __name__ == '__main__':  # pragma: no cover
    description = """A simple load tester for IRRd. Sends random !g queries."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--count', dest='count', type=int, default=5000,
                        help=f'number of queries to run (default: 5000)')
    parser.add_argument('host', type=str,
                        help='hostname of instance')
    parser.add_argument('port', type=int,
                        help='port of instance')
    args = parser.parse_args()

    main(args.host, args.port, args.count)



from twisted.internet.address import IPv4Address, UNIXAddress

from .access_check import is_client_permitted


class TestIsClientPermitted:
    peer = IPv4Address('TCP', '192.0.2.1', 99999)

    def test_no_access_list(self):
        assert is_client_permitted(self.peer, 'test.access_list', default_deny=False)
        assert not is_client_permitted(self.peer, 'test.access_list', default_deny=True)

    def test_access_list_permitted(self, config_override):
        config_override({
            'test': {
                'access_list': 'test-access-list',
            },
            'access_lists': {
                'test-access-list': ['192.0.2.0/25'],
            },
        })

        assert is_client_permitted(self.peer, 'test.access_list', default_deny=False)
        assert is_client_permitted(self.peer, 'test.access_list', default_deny=True)

    def test_access_list_denied(self, config_override):
        config_override({
            'test': {
                'access_list': 'test-access-list',
            },
            'access_lists': {
                'test-access-list': ['192.0.2.128/25'],
            },
        })

        assert not is_client_permitted(self.peer, 'test.access_list', default_deny=False)
        assert not is_client_permitted(self.peer, 'test.access_list', default_deny=True)

    def test_access_list_denied_unknown_client_address(self):
        unix_peer = UNIXAddress(b'not-supported')
        assert not is_client_permitted(unix_peer, 'test.access_list', default_deny=False)
        assert not is_client_permitted(unix_peer, 'test.access_list', default_deny=True)



#!/usr/bin/env python
# flake8: noqa: E402
import argparse
import logging
import sys

from pathlib import Path

from irrd.storage.preload import send_reload_signal

"""
Load an RPSL file into the database.
"""

logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.conf import config_init, CONFIG_PATH_DEFAULT
from irrd.mirroring.parsers import MirrorFileImportParser
from irrd.storage.database_handler import DatabaseHandler


def load(source, filename, serial, irrd_pidfile) -> int:
    dh = DatabaseHandler(enable_preload_update=False)
    dh.delete_all_rpsl_objects_with_journal(source)
    dh.disable_journaling()
    parser = MirrorFileImportParser(source, filename, serial=serial, database_handler=dh, direct_error_return=True)
    error = parser.run_import()
    if error:
        dh.rollback()
    else:
        dh.commit()
        send_reload_signal(irrd_pidfile)
    dh.close()
    if error:
        print(f'Error occurred while processing object:\n{error}')
        return 1
    return 0


def main():  # pragma: no cover
    description = """Load an RPSL file into the database."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('--irrd_pidfile', dest='irrd_pidfile', type=str, required=True,
                        help=f'path to the PID file for the running irrd instance')
    parser.add_argument('--serial', dest='serial', type=int,
                        help=f'serial number (optional)')
    parser.add_argument('--source', dest='source', type=str, required=True,
                        help=f'name of the source, e.g. NTTCOM')
    parser.add_argument('input_file', type=str,
                        help='the name of a file to read')
    args = parser.parse_args()

    config_init(args.config_file_path)

    sys.exit(load(args.source, args.input_file, args.serial, args.irrd_pidfile))


if __name__ == '__main__':  # pragma: no cover
    main()



import os

import pytest
import signal
import yaml
from typing import Dict

from . import get_setting, ConfigurationError, config_init, is_config_initialised


@pytest.fixture()
def save_yaml_config(tmpdir, monkeypatch):
    def _save(config: Dict, run_init=True):
        tmp_file = tmpdir + '/config.yaml'
        with open(tmp_file, 'w') as fh:
            fh.write(yaml.safe_dump(config))
        if run_init:
            config_init(str(tmp_file))
    return _save


class TestConfiguration:
    def test_file_not_existing(self, monkeypatch, tmpdir):
        with pytest.raises(ConfigurationError) as ce:
            config_init(str(tmpdir + '/doesnotexist.yaml'))
        assert 'Error opening config file' in str(ce)

    def test_load_invalid_yaml(self, monkeypatch, tmpdir):
        tmp_file = tmpdir + '/config.yaml'
        fh = open(tmp_file, 'w')
        fh.write('  >foo')
        fh.close()

        with pytest.raises(ConfigurationError) as ce:
            config_init(str(tmp_file))
        assert 'Error parsing YAML file' in str(ce)

    def test_load_string_file(self, save_yaml_config):
        with pytest.raises(ConfigurationError) as ce:
            save_yaml_config('foo')
        assert 'Could not find root item "irrd" in config file' in str(ce)

    def test_load_empty_config(self, save_yaml_config):
        with pytest.raises(ConfigurationError) as ce:
            save_yaml_config({})
        assert 'Could not find root item "irrd" in config file' in str(ce)

    def test_load_valid_reload_valid_config(self, monkeypatch, save_yaml_config, tmpdir, caplog):
        logfile = str(tmpdir + '/logfile.txt')
        config = {
            'irrd': {
                'database_url': 'invalid-url',
                'email': {
                    'from': 'example@example.com',
                    'smtp': '192.0.2.1'
                },
                'access-lists': {
                    'valid-list': {
                        '192/24',
                        '192.0.2.1',
                        '2001:db8::32',
                        '2001:db8::1',
                    }
                },
                'auth': {
                    'gnupg_keyring': str(tmpdir)
                },
                'sources_default': ['TESTDB2', 'TESTDB'],
                'sources': {
                    'TESTDB': {
                        'authoritative': True,
                        'keep_journal': True,
                    },
                    'TESTDB2': {
                        'nrtm_host': '192.0.2.1',
                        'nrtm_port': 43,
                        'import_serial_source': 'ftp://example.com/serial',
                        'keep_journal': True,
                    },
                },
                'log': {
                    'level': 'DEBUG',
                    'logfile_path': logfile
                },

            }
        }
        save_yaml_config(config)
        assert list(get_setting('sources_default')) == ['TESTDB2', 'TESTDB']
        assert is_config_initialised()

        config['irrd']['sources_default'] = ['TESTDB2']
        save_yaml_config(config, run_init=False)

        # Unchanged, no reload performed
        assert list(get_setting('sources_default')) == ['TESTDB2', 'TESTDB']

        os.kill(os.getpid(), signal.SIGHUP)
        assert list(get_setting('sources_default')) == ['TESTDB2']

        logfile_contents = open(logfile).read()
        assert 'Configuration successfully (re)loaded from ' in logfile_contents

    def test_load_valid_reload_invalid_config(self, save_yaml_config, tmpdir, caplog):
        save_yaml_config({
            'irrd': {
                'database_url': 'invalid-url',
                'email': {
                    'from': 'example@example.com',
                    'smtp': '192.0.2.1'
                },
                'access-lists': {
                    'valid-list': {
                        '192/24',
                        '192.0.2.1',
                        '2001:db8::32',
                        '2001:db8::1',
                    }
                },
                'auth': {
                    'gnupg_keyring': str(tmpdir)
                },
                'sources_default': ['TESTDB2', 'TESTDB'],
                'sources': {
                    'TESTDB': {
                        'authoritative': True,
                        'keep_journal': True,
                    },
                    'TESTDB2': {
                        'nrtm_host': '192.0.2.1',
                        'nrtm_port': 43,
                        'import_serial_source': 'ftp://example.com/serial',
                        'keep_journal': True,
                        'import_timer': '1234',
                    },
                },

            }
        })

        save_yaml_config({}, run_init=False)
        os.kill(os.getpid(), signal.SIGHUP)
        assert list(get_setting('sources_default')) == ['TESTDB2', 'TESTDB']
        assert 'Errors found in configuration, continuing with current settings' in caplog.text
        assert 'Could not find root item "irrd"' in caplog.text

    def test_load_invalid_config(self, save_yaml_config, tmpdir):
        config = {
            'irrd': {
                'server': {
                    'whois': {
                        'access_list': 'doesnotexist',
                    },
                    'http': {
                        'access_list': ['foo'],
                    },
                },
                'email': {
                    'footer': {'a': 1},
                },
                'access_lists': {
                    'bad-list': {
                        '192.0.2.2.1'
                    },
                },
                'sources_default': ['DOESNOTEXIST-DB'],
                'sources': {
                    'TESTDB': {
                        'keep_journal': True,
                        'import_timer': 'foo',
                        'export_timer': 'bar',
                        'nrtm_host': '192.0.2.1',
                    },
                    'TESTDB2': {
                        'authoritative': True,
                        'nrtm_host': '192.0.2.1',
                        'nrtm_port': 'not a number',
                    },
                    'TESTDB3': {
                        'authoritative': True,
                        'import_source': '192.0.2.1',
                    },
                    'lowercase': {},
                    'invalid char': {},
                },
                'log': {
                    'level': 'INVALID',
                }
            }
        }

        with pytest.raises(ConfigurationError) as ce:
            save_yaml_config(config)

        assert 'Setting database_url is required.' in str(ce)
        assert 'Setting email.from is required and must be an email address.' in str(ce)
        assert 'Setting email.smtp is required.' in str(ce)
        assert 'Setting email.footer must be a string, if defined.' in str(ce)
        assert 'Setting auth.gnupg_keyring is required.' in str(ce)
        assert 'Access lists doesnotexist referenced in settings, but not defined.' in str(ce)
        assert 'Setting server.http.access_list must be a string, if defined.' in str(ce)
        assert 'Invalid item in access list bad-list: IPv4 Address with more than 4 bytes.' in str(ce)
        assert 'Setting sources_default contains unknown sources: DOESNOTEXIST-DB' in str(ce)
        assert 'Setting keep_journal for source TESTDB can not be enabled unless either ' in str(ce)
        assert 'Setting nrtm_host for source TESTDB can not be enabled without setting import_serial_source.' in str(ce)
        assert 'Setting authoritative for source TESTDB2 can not be enabled when either nrtm_host or import_source are set.' in str(ce)
        assert 'Setting authoritative for source TESTDB3 can not be enabled when either nrtm_host or import_source are set.' in str(ce)
        assert 'Setting nrtm_port for source TESTDB2 must be a number.' in str(ce)
        assert 'Setting import_timer for source TESTDB must be a number.' in str(ce)
        assert 'Setting export_timer for source TESTDB must be a number.' in str(ce)
        assert 'Invalid source name: lowercase' in str(ce)
        assert 'Invalid source name: invalid char' in str(ce)
        assert 'Invalid log.level: INVALID' in str(ce)


class TestGetSetting:
    setting_name = 'server.whois.interface'
    env_name = 'IRRD_SERVER_WHOIS_INTERFACE'

    def test_get_setting_default(self, monkeypatch):
        monkeypatch.delenv(self.env_name, raising=False)
        assert get_setting(self.setting_name) == '::0'

    def test_get_setting_env(self, monkeypatch):
        monkeypatch.setenv(self.env_name, 'env_value')
        assert get_setting(self.setting_name) == 'env_value'



import logging
from twisted.web import server, resource

from .request_handlers import DatabaseStatusRequest
from ..access_check import is_client_permitted

logger = logging.getLogger(__name__)


class DatabaseStatusResource(resource.Resource):
    """
    A twisted HTTP server resource for the status of the database.
    """
    isLeaf = True  # noqa: N815

    def render_GET(self, request: server.Request):  # noqa: N802
        """Render the database status in plain text."""
        if not self.is_client_permitted(request):
            request.setResponseCode(403)
            return b'Access denied'

        result_txt = DatabaseStatusRequest().generate_status()
        request.setHeader(b'Content-Type', b'text/plain; charset=utf-8')
        return result_txt.encode('utf-8')

    def is_client_permitted(self, request) -> bool:
        """
        Determine whether a client is permitted to access this interface,
        based on the server.http.access_list setting.
        """
        return is_client_permitted(request.getClientAddress(), 'server.http.access_list')


root = resource.Resource()
v1 = resource.Resource()
root.putChild(b'v1', v1)
v1.putChild(b'status', DatabaseStatusResource())

http_site = server.Site(root)
http_site.displayTracebacks = False
